import cv2
import numpy as np
from vhh_cmc.Configuration import Configuration
from vhh_cmc.PreProcessing import PreProcessing
import os
from vhh_cmc.OpticalFlow import OpticalFlow
from vhh_cmc.Video import Video

#import matplotlib
#matplotlib.use('Qt5Agg')

class CMC(object):
    """
    Main class of camera movements classification (cmc) package.
    """
    def __init__(self, config_file: str):
        """
        Constructor

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                                       must be with extension ".yaml"
        """
        print("create instance of cmc ... ")

        if (config_file == ""):
            print("No configuration file specified!")
            exit()

        self.config_instance = Configuration(config_file)
        self.config_instance.loadConfig()

        if (self.config_instance.debug_flag == True):
            print("DEBUG MODE activated!")
            self.debug_results = self.config_instance.path_debug_results
            print("save debug results to: " + str(self.debug_results))

        self.pre_processing_instance = PreProcessing(config_instance=self.config_instance)

    def runOnSingleVideo(self, shots_per_vid_np=None, max_recall_id=-1):
        """
        Method to run cmc classification on specified video.

        :param shots_per_vid_np: [required] numpy array representing all detected shots in a video
                                 (e.g. sid | movie_name | start | end )
        :param max_recall_id: [required] integer value holding unique video id from VHH MMSI system
        """

        print("run cmc classifier on single video ... ")

        if (type(shots_per_vid_np) == None):
            print("ERROR: you have to set the parameter shots_per_vid_np!")
            exit()

        if (max_recall_id == -1 or max_recall_id == 0):
            print("ERROR: you have to set a valid max_recall_id [1-n]!")
            exit()

        if (self.config_instance.debug_flag == True):
            # load shot list from result file
            shots_np = self.loadSbdResults(self.config_instance.sbd_results_path)
            debug_sid = 125 # 33 12 14 4 #tilt ids: 86  104 73 108 125  pan ids: 5 7 21 13 28 127 na ids: 3 24 77
        else:
            shots_np = shots_per_vid_np
            debug_sid = -1

        if (len(shots_np) == 0):
            print("ERROR: there must be at least one shot in the list!")
            exit()

        # read all frames of video
        vid_name = shots_np[0][0]       

        '''
        # map vid name with origianl file names for icprs paper
        orig_file_names = os.listdir(self.config_instance.path_videos)
        print(orig_file_names)
        for orig_name in orig_file_names:
            if(vid_name.split('.')[0] + "_(" in orig_name):
                vid_name = orig_name
        print(vid_name)
        '''

        if(self.config_instance.save_eval_results == 1):
            print("Evaluation mode is activated ...")
            print(self.config_instance.path_videos + "/" + vid_name)
            #cap = cv2.VideoCapture(self.config_instance.path_videos + "/" + vid_name)

            vid_instance = Video()
            vid_instance.load(self.config_instance.path_videos + "/" + vid_name)
        else:
            print(self.config_instance.path_videos + "/" + vid_name)
            #cap = cv2.VideoCapture(self.config_instance.path_videos + "/" + vid_name)

            vid_instance = Video()
            vid_instance.load(self.config_instance.path_videos + "/" + vid_name)

            #vid_name = "20_3776.mp4"  #training_data/pan/69_24604.mp4;0;0;142  training_data/pan/77_66460.mp4;0;0;124 training_data/tilt/20_3776.mp4;1;0;125
            #vid_instance.load("/data/share/datasets/cmc_final_dataset_v2/training_data/tilt" + "/" + vid_name)
            #shots_np = np.array([["999", "1", "0", "125"]])

        print(shots_np)
        results_cmc_l = []
        results_seq_cmc_l = []
        for data in vid_instance.getFramesByShots(shots_np, preprocess=self.pre_processing_instance.applyTransformOnImg):
            frames_per_shots_np = data['Images']
            shot_id = data['sid']
            vid_name = data['video_name']
            start = data['start']
            stop = data['end']

            if(shot_id != debug_sid) and (self.config_instance.debug_flag == True):
                continue

            print(f'sid: {shot_id}')
            print(f'vid_name: {vid_name}')
            print(f'frames_per_shot: {frames_per_shots_np.shape}')
            print(f'start: {start}, end: {stop}')

            shot_len = stop - start
            MIN_NUMBER_OF_FRAMES_PER_SHOT = 5 # 10
            if(shot_len <= MIN_NUMBER_OF_FRAMES_PER_SHOT ):
                #print("shot length is too small!")
                class_name = "NA"
                seq_dict_l = []
            else:
                # add new optical flow version
                optical_flow_instance = OpticalFlow(video_frames=frames_per_shots_np,
                                                    algorithm=None,
                                                    config_instance=self.config_instance)
                class_name, seq_dict_l = optical_flow_instance.runDense(vid_name, shot_id, start, stop)
            
            # prepare results
            print(str(vid_name) + ";" + str(shot_id) + ";" + str(start) + ";" + str(stop) + ";" + str(class_name))
            results_cmc_l.append([str(vid_name) + ";" + str(shot_id) + ";" + str(start) + ";" + str(stop) + ";" + str(class_name)])

            results_seq_cmc_l.extend(seq_dict_l)

            if (shot_id == debug_sid and self.config_instance.debug_flag == True):
                break

        results_cmc_np = np.array(results_cmc_l)

        if(self.config_instance.debug_flag == True):
            print(results_cmc_np)
            
        # export results
        if (self.config_instance.save_eval_results == 1):
            self.exportCmcResults(vid_name, results_cmc_np)
        else:
            #self.exportCmcResults(str(max_recall_id), results_cmc_np)

            # sequence-based results
            import json
            with open(self.config_instance.path_final_results + "/" + str(vid_name.split('.')[0] + "." + vid_name.split('.')[1]) + ".json", 'w') as fp:
                json.dump(results_seq_cmc_l, fp)

    def loadSbdResults(self, sbd_results_path):
        """
        Method for loading shot boundary detection results as numpy array

        .. note::
            Only used in debug_mode.

        :param sbd_results_path: [required] path to results file of shot boundary detection module (vhh_sbd)
        :return: numpy array holding list of detected shots.
        """

        # open sbd results
        fp = open(sbd_results_path, 'r')
        lines = fp.readlines()
        lines = lines[1:]

        lines_n = []
        for i in range(0, len(lines)):
            line = lines[i].replace('\n', '')
            line_split = line.split(';')
            lines_n.append([line_split[0], os.path.join(line_split[1]), line_split[2], line_split[3]])
        lines_np = np.array(lines_n)
        # print(lines_np.shape)

        return lines_np

    def loadStcResultsFromJson(self, stc_results_path):
        """
        tbd

        .. note::
            Only used in debug_mode.

        :param stc_results_path: [required] path to results file of shot boundary detection module (vhh_sbd)
        :return: numpy array holding list of detected shots.
        """
        
        vid_name = stc_results_path.split('/')[-1]
        #print(vid_name)
        vid_name = vid_name.split('.')[0]
        #print(vid_name)
        vid_name = vid_name.split('-')[0]
       
        import json
        with open(stc_results_path) as f:
            shots_per_vid = json.load(f)

        lines_n = []
        for shot in shots_per_vid:
            #print(shot)
            lines_n.append([vid_name + ".m4v", shot['shotId'], shot['inPoint'], shot['outPoint'], shot['shotType']])
        lines_np = np.array(lines_n)
        #print(lines_np)
        return lines_np

    def loadCmcResultsFromJson(self, cmc_results_path):
        """
        Method for loading camera movments results as numpy array

        .. note::
            Only used in debug_mode.

        :param cmc_results_path: [required] path to results file of shot boundary detection module (vhh_sbd)
        :return: numpy array holding list of detected shots.
        """

        vid_name = cmc_results_path.split('/')[-1]
        vid_name = vid_name.split('.')[0] + "." + vid_name.split('.')[1]
        print(vid_name)
        #exit()

        import json
        with open(cmc_results_path) as f:
            shots_per_vid = json.load(f)

        lines_n = []
        for shot in shots_per_vid:
            #print(shot)
            lines_n.append([vid_name + ".m4v", shot['shotId'], shot['start'], shot['stop'], shot['cmcType']])
        lines_np = np.array(lines_n)
        #print(lines_np)
        return lines_np

    def loadCmcResults(self, cmc_results_path):
        """
        Method for loading camera movments results as numpy array

        .. note::
            Only used in debug_mode.

        :param cmc_results_path: [required] path to results file of shot boundary detection module (vhh_sbd)
        :return: numpy array holding list of detected shots.
        """
        #print(cmc_results_path)
        #exit()

        # open sbd results
        fp = open(cmc_results_path, 'r')
        lines = fp.readlines()
        lines = lines[1:]

        lines_n = []
        for i in range(0, len(lines)):
            line = lines[i].replace('\n', '')
            line_split = line.split(';')
            lines_n.append([line_split[0], os.path.join(line_split[1]), line_split[2], line_split[3], line_split[4]])
        lines_np = np.array(lines_n)
        # print(lines_np.shape)

        return lines_np

    def export_cmc_sequences_as_file(self, seq_np, dst_path="./vhh_mmsi_eval_db_tiny/shots/"):
        print("export shot as video")

        print(seq_np.shape)
        
        vid_name = seq_np[0][0]
        # map vid name with origianl file names for icprs paper
        orig_file_names = os.listdir(self.config_instance.path_videos)
        print(orig_file_names)
        for orig_name in orig_file_names:
            if(vid_name.split('.')[0] + "_(" in orig_name):
                vid_name = orig_name
        ''''''
        print(vid_name)
       

        vid_instance = Video()
        vid_instance.load(self.config_instance.path_videos + "/" + vid_name)

        h = int(vid_instance.height)
        w = int(vid_instance.width)
        fps = int(vid_instance.frame_rate)

        print(h)
        print(w)
        print(fps)

        for i, data in enumerate(vid_instance.getFramesByShots(seq_np)):
            frames_per_shots_np = data['Images']
            shot_id = data['sid']
            #vid_name = data['video_name']
            start = data['start']
            stop = data['end']
            cmc_class = data['cmc_class']

            print("######################")
            print(i)
            print(i % 32 == 0)
            print(f'sid: {shot_id}')
            #print(f'vid_name: {vid_name}')
            print(f'frames_per_shot: {frames_per_shots_np.shape}')
            print(f'start: {start}, end: {stop}')
            print(f'cmc_class: {cmc_class}')
            file_name = "vid_" + str(vid_name.split('_')[0]) + "_sid_" + str(shot_id) + "_start_" + str(start) + "_stop_" + str(stop) + "_classname_" + str(cmc_class)
            #"vid_xxx_sid_xxx_start_yyy_stop_xxxx_classname_XXX"

            print("save video! ")
            fourcc = cv2.VideoWriter_fourcc(*'MJPG')
            out = cv2.VideoWriter(dst_path + "/" + file_name + ".avi", fourcc, 12, (w, h))
            
            '''
            font = cv2.FONT_HERSHEY_SIMPLEX
            bottomLeftCornerOfText = (30, 50)
            fontScale = 1
            fontColor = (255, 255, 255)
            lineType = 2
            '''
            for j in range(0, len(frames_per_shots_np)):
                frame = frames_per_shots_np[j]
                '''
                cv2.rectangle(frame, (0, 0), (350, 80), (0, 0, 255), -1)
                cv2.putText(frame, "CameraMovement-Type: " + cmc_class,
                            bottomLeftCornerOfText,
                            font,
                            fontScale,
                            fontColor,
                            lineType)
                '''
                out.write(frame)
                #cv2.imshow("test", frame)
                #k = cv2.waitKey(10)
            
            out.release()

    def exportCmcResults(self, fName, cmc_results_np: np.ndarray):
        """
        Method to export cmc results as csv file.

        :param fName: [required] name of result file.
        :param cmc_results_np: numpy array holding the camera movements classification predictions for each shot of a movie.
        """

        print("export results to csv!")

        if (len(cmc_results_np) == 0):
            print("ERROR: numpy is empty")
            exit()

        # open stc resutls file
        if (self.config_instance.debug_flag == True):
            fp = open(self.debug_results + "/" + fName + ".csv", 'w')
        elif (self.config_instance.save_eval_results == 1):

            print(self.config_instance.path_eval_results)
            fp = open(self.config_instance.path_eval_results + "/" + fName.split('/')[-1].split('.')[0] + ".csv", 'w')
        else:
            fp = open(self.config_instance.path_final_results + "/" + fName + ".csv", 'w')



        header = "vid_name;shot_id;start;end;cmc"
        fp.write(header + "\n")

        for i in range(0, len(cmc_results_np)):
            tmp_line = str(cmc_results_np[i][0])
            for c in range(1, len(cmc_results_np[i])):
                tmp_line = tmp_line + ";" + cmc_results_np[i][c]
            fp.write(tmp_line + "\n")

        fp.close()

    def export_shots_as_file(self, shots_np, dst_path="./vhh_mmsi_eval_db_tiny/shots/", max_num_of_shots=10, VIS_TEXT_FLAG=False):
        print("export shot as video")

        print(shots_np.shape)

        vid_name = shots_np[0][0]
        vid_instance = Video()
        vid_instance.load(self.config_instance.path_videos + "/" + vid_name)

        h = int(vid_instance.height)
        w = int(vid_instance.width)
        fps = int(vid_instance.frame_rate)

        print(h)
        print(w)
        print(fps)

        shot_cnt = 0
        for i, data in enumerate(vid_instance.getFramesByShots(shots_np, preprocess=None)):
            frames_per_shots_np = data['Images']
            shot_id = data['sid']
            vid_name = data['video_name']
            start = data['start']
            stop = data['end']
            camera_movement_class = data["cmc_class"]

            print("######################")
            print(i)
            print(i % 32 == 0)
            print(f'sid: {shot_id}')
            print(f'vid_name: {vid_name}')
            print(f'frames_per_shot: {frames_per_shots_np.shape}')
            print(f'start: {start}, end: {stop}')
            print(f'camera_movement_class: {camera_movement_class}')

            #if (camera_movement_class == "PAN" or camera_movement_class == "TILT" or camera_movement_class == "pan" or camera_movement_class == "tilt"):
            if (camera_movement_class == "NA" or camera_movement_class == "na"):
                print("save video! ")

                if (shot_cnt >= max_num_of_shots and max_num_of_shots != -1):
                    break

                fourcc = cv2.VideoWriter_fourcc(*'MJPG')
                out = cv2.VideoWriter(dst_path + "/vid_" + str(vid_name.split('.')[0]) + "_sid_" + str(shot_id) + "_" + camera_movement_class + "_" + str(i) + ".avi", fourcc, 12, (w, h))

                font = cv2.FONT_HERSHEY_SIMPLEX
                bottomLeftCornerOfText = (30, 50)
                fontScale = 1
                fontColor = (255, 255, 255)
                lineType = 2

                for j in range(0, len(frames_per_shots_np)):
                    frame = frames_per_shots_np[j]
                    if(VIS_TEXT_FLAG == True):
                        cv2.rectangle(frame, (0, 0), (450, 80), (0, 0, 255), -1)
                        cv2.putText(frame, "Camera-Movement: " + camera_movement_class,
                                    bottomLeftCornerOfText,
                                    font,
                                    fontScale,
                                    fontColor,
                                    lineType)
                    out.write(frame)

                out.release()

                shot_cnt = shot_cnt + 1







from sys import exit


# config_path ... the path to the config file
def parse_config(config_path="config.yaml"):
    from yaml import load, CLoader

    try:
        with open(config_path) as stream:
            config = load(stream=stream, Loader=CLoader)
            return config
    except ImportError:
        print("failing to load config.yaml.")
        exit(2)


# argv ... the system arguments
# defaults ... a dictionary with default options
def parse_sys_arg(argv, defaults):
    from getopt import getopt, GetoptError

    short_arguments = "hs:p:b:n:a:m:i:o:g:c:d:t:f:l:"

    long_arguments = ['SENSITIVITY=', 'SPECIFICITY=', 'BORDER=', 'NUMBER_OF_FEATURES=', 'ANGLE_DIFF_LIMIT=', 'MODE=',
                      'INPUT_VIDEO=', 'OUTPUT_VIDEO=', 'OUTPUT_PLOT=', 'OUTPUT_CSV=', 'INPUT_PATH=', 'OUTPUT_PATH=',
                      'BEGIN_FRAME=', 'END_FRAME=']

    helper_string = "\n".join(["<your_script_name.py>", "-s <SENSITIVITY>", "-p <SPECIFICITY>", "-b <BORDER>",
                               "-n <NUMBER_OF_FEATURES>", "-a <ANGLE_DIFF_LIMIT>", "-m <MODE>", "-i <INPUT_VIDEO>",
                               "-o <OUTPUT_VIDEO>", "-g <OUTPUT_PLOT>", "-c <OUTPUT_CSV>", "-d <INPUT_PATH>",
                               "-t <OUTPUT_PATH>", "-f <BEGIN_FRAME>", "-l <END_FRAME>"])

    try:
        opts, args = getopt(argv, short_arguments, long_arguments)

        for opt, arg in opts:

            if opt == "-h":
                print(helper_string)
                exit()

            for i in range(0, len(long_arguments)):

                short_argument = short_arguments[1 + i * 2]
                long_argument = long_arguments[i].split("=")[0]

                if opt in ("-" + short_argument, "--" + long_argument):
                    defaults[long_argument] = arg
                    print("updating default ", long_argument, " to ", arg)

        return defaults

    except GetoptError:
        print(helper_string)
        exit(2)

class KeyHandler:

    # funcs is a dictionary { keycode: function, keycode: function, ... }
    def __init__(self, delay, funcs):
        self.delay = delay
        self.funcs = funcs

    def wait_and_handle_key(self):
        import cv2
        c = cv2.waitKey(self.delay) & 0xFF
        for k, v in self.funcs.items():
            if c == k:
                v()


from matplotlib import pyplot as plt
import numpy as np


# method to visualize scatter plots
def visualizeScatterPlot(x, y, y_max, y_min):
    # plot angles over time
    plt.figure(1)
    plt.plot(np.arange(y), x)
    plt.ylim(ymax=y_max, ymin=y_min)
    plt.grid(True)
    plt.show()
    #plt.draw()
    #plt.pause(0.02)


# method to visualize hist plots


# method to visualize imageplots


# method to visualize configurable subplots





import cv2
import numpy as np
import argparse
from matplotlib import pyplot as plt


class OpticalFlow_SIFT(object):
    def __init__(self, video_frames=None):
        #self.video_name = "C:\\Users\\dhelm\\Documents\\slow_traffic_small.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\tilt\\b88f0e71-a0f2-4efe-ae0d-5b83a0770b73_32.mp4"  # tilt unten nach oben
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\tilt\\35_25178.mp4"  # tilt oben nach unten
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\tilt\\94_24357.mp4" # tilt unten nach oben
        #1371a561-6b19-4d69-8210-1347ca75eb90_96
        # self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\1371a561-6b19-4d69-8210-1347ca75eb90_104.mp4"
        # self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\tilt\\tilt_130_74088.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\130_74173.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\1371a561-6b19-4d69-8210-1347ca75eb90_96.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\066f929d-6434-4ea9-844b-e066f57b6c28_99.mp4" # rechts nach links
        #self.video_name = "C:\\Users\\dhelm\\Documents\\999.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\444.mp4"
        #self.video_name = "C:\\Users\\dhelm\\Documents\\training_data_patrick_link\\training_data\\pan\\72_65942.mp4"  # pan links nach rechts

        self.video_frames = video_frames

    def getMatches(self, frame1, frame2):
        kp_curr_list = []
        kp_prev_list = []

        sift = cv2.xfeatures2d.SIFT_create()
        kp_prev, descriptor_prev = sift.detectAndCompute(frame1, None)
        kp_curr, descriptor_curr = sift.detectAndCompute(frame2, None)

        #print(len(kp_prev))
        #print(len(kp_curr))
        #print(len(descriptor_prev))
        #print(len(descriptor_curr))
        #print(descriptor_prev)
        #print(descriptor_curr)

        #print(type(descriptor_prev))
        #print(type(descriptor_curr))

        out_frame1 = frame1.copy()
        #img = cv2.drawKeypoints(old_frame, kp, out_img)
        #out_img_prev = cv2.drawKeypoints(frames_np[i-1],
        #                                 kp_prev,
        #                                 out_img_prev,
        #                                 flags=cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)
        out_frame2 = frame2.copy()
        # img = cv2.drawKeypoints(old_frame, kp, out_img)
        #out_img_curr = cv2.drawKeypoints(frames_np[i - 1],
        #                                 kp_curr,
        #                                 out_img_curr,
        #                                 flags=cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)
        #print(out_img_prev.shape)
        #print(out_img_curr.shape)

        # Create a Brute Force Matcher object.
        bf = cv2.BFMatcher()

        # Perform the matching between the ORB descriptors of the training image and the test image
        try:
            matches = bf.knnMatch(descriptor_prev, descriptor_curr, k=2)
            #print(matches)
        except:
            return kp_prev_list, kp_curr_list


        # Apply ratio test
        good_matches = []
        if(len(matches) > 0):
            if(len(matches[0]) == 2 ):
                for m, n in matches:
                    if m.distance < 0.25 * n.distance:
                        good_matches.append([m])
            elif(len(matches[0]) == 1):
                for m in matches:
                    if m[0].distance < 0.25:
                        good_matches.append([m])
        #good_matches = np.squeeze(np.array(good_matches)).tolist()
        #print((good_matches))

        #print(matches)
        #print(type(good_matches))
        if (len(good_matches) == 0):
            return kp_prev_list, kp_curr_list


        #print(descriptor_prev.shape)
        #matches = bf.knnMatch(descriptor_prev, descriptor_curr, k=3)

        ## The matches with shorter distance are the ones we want.
        #good_matches = sorted(good_matches, key=lambda x: x.distance)

        for match in good_matches:
            kp_curr_list.append(kp_curr[match[0].trainIdx].pt)
            kp_prev_list.append(kp_prev[match[0].queryIdx].pt)

        '''
        result = cv2.drawMatches(out_img_curr, kp_curr, out_img_prev, kp_prev, matches[:1], out_img_prev, flags=2)
        # Display the best matching points
        plt.title('Best Matching Points')
        plt.imshow(result)
        plt.draw()
        plt.pause(0.05)
        '''
        return kp_prev_list, kp_curr_list



import numpy as np
import cv2
import datetime
from PIL import Image


class Video(object):
    """
    This class is representing a video. Each instance of this class is holding the properties of one Video.
    """

    def __init__(self):
        """
        Constructor
        """

        # printCustom("create instance of video class ... ", STDOUT_TYPE.INFO);
        self.vidFile = ''
        self.vidName = ""
        self.frame_rate = 0
        self.channels = 0
        self.height = 0
        self.width = 0
        self.format = ''
        self.length = 0
        self.number_of_frames = 0
        self.vid = None
        self.convert_to_gray = False
        self.convert_to_hsv = False

    def load(self, vidFile: str):
        """
        Method to load video file.
        :param vidFile: [required] string representing path to video file
        """

        # print(vidFile)
        # printCustom("load video information ... ", STDOUT_TYPE.INFO);
        self.vidFile = vidFile;
        if (self.vidFile == ""):
            # print("A")
            print("ERROR: you must add a video file path!");
            exit(1);
        self.vidName = self.vidFile.split('/')[-1]

        print(self.vidFile)
        self.vid = cv2.VideoCapture(self.vidFile);

        if (self.vid.isOpened() == False):
            # print("B")
            print("ERROR: not able to open video file!");
            exit(1);

        status, frm = self.vid.read();

        self.channels = frm.shape[2];
        self.height = self.vid.get(cv2.CAP_PROP_FRAME_HEIGHT);
        self.width = self.vid.get(cv2.CAP_PROP_FRAME_WIDTH);
        self.frame_rate = self.vid.get(cv2.CAP_PROP_FPS);
        self.format = self.vid.get(cv2.CAP_PROP_FORMAT);
        self.number_of_frames = self.vid.get(cv2.CAP_PROP_FRAME_COUNT);

        self.vid.release();

    def printVIDInfo(self):
        """
        Method to a print summary of video properties.
        """

        print("---------------------------------");
        print("Video information");
        print("filename: " + str(self.vidFile));
        print("format: " + str(self.format));
        print("fps: " + str(self.frame_rate));
        print("channels: " + str(self.channels));
        print("width: " + str(self.width));
        print("height: " + str(self.height));
        print("nFrames: " + str(self.number_of_frames));
        print("---------------------------------");

    def getFrame(self, frame_id):
        """
        Method to get one frame of a video on a specified position.
        :param frame_id: [required] integer value with valid frame index
        :return: numpy frame (WxHx3)
        """

        self.vid.open(self.vidFile)
        if (frame_id >= self.number_of_frames):
            print("ERROR: frame idx out of range!")
            return []

        # print("Read frame with id: " + str(frame_id));
        time_stamp_start = datetime.datetime.now().timestamp()

        self.vid.set(cv2.CAP_PROP_POS_FRAMES, frame_id)
        status, frame_np = self.vid.read()
        self.vid.release()

        if (status == True):
            if (self.convert_to_gray == True):
                frame_np = cv2.cvtColor(frame_np, cv2.COLOR_BGR2GRAY)
                # print(frame_gray_np.shape);
            if (self.convert_to_hsv == True):
                frame_np = cv2.cvtColor(frame_np, cv2.COLOR_BGR2HSV)
                h, s, v = cv2.split(frame_np)

        time_stamp_end = datetime.datetime.now().timestamp()
        time_diff = time_stamp_end - time_stamp_start
        # print("time: " + str(round(time_diff, 4)) + " sec");

        return frame_np

    # Returns Video Shot by Shot
    def getFramesByShots(self, shots_np, preprocess=None):

        # initialize video capture
        cap = cv2.VideoCapture(self.vidFile)

        frame_number = 0
        for i in range(0, len(shots_np)):
            shot = shots_np[i]

            frame_l = []
            frames_orig = []

            sid = int(shot[1])
            start_idx = int(shot[2])
            stop_idx = int(shot[3])
            #cmc_class = shot[4]

            # print(f"Retrieving Frames for Shot {sid} (frames {frame_number} to {stop_idx})...")
            while frame_number <= stop_idx:
                # read next frame
                success, image = cap.read()
                frame_number = frame_number + 1
                #print(frame_number)

                # if(start_idx == stop_idx):
                #    cv2.imshow("frame", image)
                #    k = cv2.waitKey()

                # skip to start position (for gradual cuts)
                if frame_number < start_idx:
                    # print(frame_number)
                    continue

                if success == True:
                    # if ( (frame_number >= start_idx and frame_number <= stop_idx) or (start_idx == stop_idx) ):
                    if (preprocess != None):
                        image_pre = preprocess(image)
                        frame_l.append(image_pre)
                    else:
                        frame_l.append(image)
                else:
                    break

            yield {"Images": np.array(frame_l),
                   "sid": sid,
                   "video_name": self.vidName,
                   "start": start_idx,
                   "end": stop_idx,
                   #"cmc_class": cmc_class
                   }
        cap.release()


from vhh_cmc.Configuration import Configuration
from vhh_cmc.PreProcessing import PreProcessing
import os
import numpy as np
import cv2

from sklearn.metrics import classification_report
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import f1_score
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics import confusion_matrix


class Evaluation(object):
    """
    Evaluation class includes all methods to evaluate the implemented algorithms.
    """

    def __init__(self, config_instance: Configuration):
        """
        Constructor.

        :param config_instance: object instance of type Configuration
        """
        print("create instance of class evaluation ... ")

        self.config_instance = config_instance
        self.path_eval_dataset = self.config_instance.path_eval_dataset
        self.path_gt_data = self.config_instance.path_gt_data
        self.path_eval_results = self.config_instance.path_eval_results

        self.pre_processing_instance = PreProcessing(self.config_instance)

        self.all_shot_file_list = []
        self.final_dataset_np = []

    def load_cmc_eval_db_v2(self):
        """
        This method is used to load the dataset used to evaluate the algorithm.
        The dataset must have the following structure:
        dataset_root_dir/training_data/
        dataset_root_dir/training_data/tilt/
        dataset_root_dir/training_data/pan/
        dataset_root_dir/training_data/annotation/xxx.flist
        """
        print("load eval dataset ... ")

        # load samples
        tilt_samples_path = self.path_eval_dataset + "/training_data/tilt/"
        pan_samples_path = self.path_eval_dataset + "/training_data/pan/"
        na_samples_path = self.path_eval_dataset + "/training_data/na/"
        path_annotations = self.path_eval_dataset + "/annotation_new/"

        tilt_shot_file_list = os.listdir(tilt_samples_path)
        pan_shot_file_list = os.listdir(pan_samples_path)
        na_shot_file_list = os.listdir(na_samples_path)

        self.all_shot_file_list = tilt_shot_file_list + pan_shot_file_list + na_shot_file_list
        self.all_shot_file_list.sort()
        all_shot_file_np = np.array(self.all_shot_file_list)
        #print(len(self.all_shot_file_list))

        # load groundtruth labels
        #test_gt_labels_file = path_annotations + "/annotations_tiny.csv"
        test_gt_labels_file = path_annotations + "/annotations.csv"
        #print(test_gt_labels_file)
      
        fp = open(test_gt_labels_file, 'r')
        lines = fp.readlines()
        fp.close()

        gt_annotation_list = []
        for line in lines:
            #print(line)
            line = line.replace('\n', '')
            line = line.replace('\\', '/')
            line = line.replace('\ufeff', '')
            line_split = line.split(';')
            gt_annotation_list.append([line_split[0], line_split[2], line_split[3]])
        gt_annotation_np = np.array(gt_annotation_list)

        final_dataset = []
        for i in range(0, len(gt_annotation_np)):
            #path = os.path.join(self.path_eval_dataset, gt_annotation_np[i][0])
            path = gt_annotation_np[i][0]
            video_name = path.split('/')[-1]
            start = int(gt_annotation_np[i][1])
            stop = int(gt_annotation_np[i][2]) - 1
            class_name = gt_annotation_np[i][0].split('/')[1]

            #print(video_name)
            idx = np.squeeze(np.where(video_name == all_shot_file_np)[0])
            sample_path = all_shot_file_np[idx]
            final_dataset.append([path, i, start, stop, class_name])

        #final_dataset.sort()
        self.final_dataset_np = np.array(final_dataset)
        print(self.final_dataset_np)

    def load_vhhmmsi_GT_V2_db(self):
        """
        This method is used to load the dataset used to evaluate the algorithm.
        The dataset must have the following structure:
        dataset_root_dir/training_data/
        dataset_root_dir/training_data/tilt/
        dataset_root_dir/training_data/pan/
        dataset_root_dir/training_data/annotation/xxx.flist
        """
        print("load eval dataset ... ")

        path_annotations = self.path_eval_dataset + "/annotations/cmc/"

        # load groundtruth labels
        test_gt_labels_file = os.listdir(path_annotations)
        test_gt_labels_file.sort()
        print(test_gt_labels_file)
        #test_gt_labels_file = test_gt_labels_file[1:2]
        #print(test_gt_labels_file)

        final_dataset = []
        gt_annotation_list = []
        for file in test_gt_labels_file:
            fp = open(path_annotations + file, 'r')
            lines = fp.readlines()
            fp.close()
            lines = lines[1:]
            
            for line in lines:
                #print(line)
                line = line.replace('\n', '')
                line = line.replace('\\', '/')
                line = line.replace('\ufeff', '')
                line_split = line.split(';')
                #print(line_split)
                #if(line_split[4] == "track"):
                #    #line_split[4] = "na"
                #    continue

                gt_annotation_list.append([line_split[0], int(line_split[1]), int(line_split[2]), int(line_split[3]), line_split[4]])
                
        gt_annotation_np = np.array(gt_annotation_list)
        self.final_dataset_np = gt_annotation_np

    def run_evaluation(self, idx=None):
        """
        This method is used to start and run the evaluation process.
        """
        print("calculate evaluation metrics ... ")

        if (idx != None):
            results_file_list = [f for f in os.listdir(self.path_eval_results) if f.endswith('.csv')][idx:idx+1]
            self.final_dataset_np = self.final_dataset_np[idx:idx+1]
            #print(self.final_dataset_np)
            #exit()
        else:
            # load all predictions and merge
            results_file_list = [f for f in os.listdir(self.path_eval_results) if f.endswith('.csv')]
            results_file_list.sort()
            print(results_file_list)

            #print(self.path_eval_results)
            #exit()

            final_gt_np = self.final_dataset_np
            print(self.final_dataset_np[:20])
            idx_track = np.where(self.final_dataset_np[:, 4:] == "track")[0]
            print(idx_track)
            #exit()

        pred_list = []
        pred_np = []
        for i, file in enumerate(results_file_list):
            results_file = os.path.join(self.path_eval_results, file)
            #print(results_file)
            fp = open(results_file, 'r')
            lines = fp.readlines()
            lines = lines[1:]
            lines.sort()
            fp.close()

            for line in lines:
                line = line.replace('\n', '')
                line_split = line.split(';')
                pred_list.append([line_split[0], int(line_split[1]), int(line_split[2]), int(line_split[3]), line_split[4]])
        pred_np = np.array(pred_list)

        # add sorting part
        pred_sort_np = pred_np[pred_np[:, 1].argsort()]
        final_dataset_sort_np = self.final_dataset_np[self.final_dataset_np[:, 1].argsort()]

        # remove paths
        for i in range(0, len(final_dataset_sort_np)):
            name = final_dataset_sort_np[i][0].split('/')[-1]
            final_dataset_sort_np[i][0] = name

        pred_np_prep = np.squeeze(pred_sort_np[:, 4:])
        gt_np_prep = np.squeeze(final_dataset_sort_np[:, 4:])

        accuracy, precision, recall, f1_score = self.calculate_metrics(y_score=pred_np_prep, y_test=gt_np_prep)
        return accuracy, precision, recall, f1_score

    def calculate_metrics(self, y_score, y_test):
        """
        This method is used to calculate the metrics: precision, recall, f1score.
        Furthermore, the confusion matrix is generated and stored as figure on a specified location.

        :param y_score: This parameter must hold a valid numpy array with the class prediction per shot .
        :param y_test: This parameter must hold a valid numpy array with the groundtruth labels per shot.
        """
        #print(len(y_score))
        #print(len(y_test))

        y_score = [s.lower() for s in y_score]
        y_test = [s.lower() for s in y_test]

        self.config_instance.class_names = [s.lower() for s in self.config_instance.class_names]

        #print(y_score)
        #print(y_test)
        print(self.config_instance.class_names)

        # accuracy: (tp + tn) / (p + n)
        accuracy = accuracy_score(y_test, y_score)
        print('Accuracy: %f' % accuracy)
        # precision tp / (tp + fp)
        precision = precision_score(y_test, y_score, average='macro')
        print('Precision: %f' % precision)
        # recall: tp / (tp + fn)
        recall = recall_score(y_test, y_score, average='macro')
        print('Recall: %f' % recall)
        # f1: 2 tp / (2 tp + fp + fn)
        f1 = f1_score(y_test, y_score, average='macro')
        print('F1 score: %f' % f1)

        # confusion matrix
        matrix = confusion_matrix(y_test, y_score, labels=self.config_instance.class_names)
        print(matrix)
        print(classification_report(y_test, y_score))

        print("save confusion matrix ...")
        self.plot_confusion_matrix(cm=matrix,
                                   target_names=self.config_instance.class_names,
                                   title='Confusion matrix',
                                   cmap=None,
                                   normalize=True,
                                   path=self.config_instance.path_eval_results + "/confusion_matrix_normalize.pdf")
        self.plot_confusion_matrix(cm=matrix,
                                   target_names=self.config_instance.class_names,
                                   title='Confusion matrix',
                                   cmap=None,
                                   normalize=False,
                                   path=self.config_instance.path_eval_results + "/confusion_matrix.pdf")
        ''''''

        return accuracy, precision, recall, f1

    def plot_confusion_matrix(self, cm=None,
                          target_names=[],
                          title='Confusion matrix',
                          cmap=None,
                          normalize=True,
                          path=""):
        import matplotlib.pyplot as plt
        import numpy as np
        import itertools
        import matplotlib
        matplotlib.use('Agg')
        from matplotlib import pyplot as plt

        plt.rc('pdf', fonttype=42)


        accuracy = np.trace(cm) / float(np.sum(cm))
        misclass = 1 - accuracy

        if cmap is None:
            cmap = plt.get_cmap('Blues')

        plt.figure() #figsize=(8, 6)
        plt.imshow(cm, interpolation='nearest', cmap=cmap)
        #plt.title(title)
        plt.colorbar()

        if target_names is not None:
            tick_marks = np.arange(len(target_names))
            plt.xticks(tick_marks, target_names, rotation=45)
            plt.yticks(tick_marks, target_names)

        if normalize:
            cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

        thresh = cm.max() / 1.5 if normalize else cm.max() / 2
        for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
            if normalize:
                plt.text(j, i, "{:0.4f}".format(cm[i, j]),
                        horizontalalignment="center",
                        color="white" if cm[i, j] > thresh else "black")
            else:
                plt.text(j, i, "{:,}".format(cm[i, j]),
                        horizontalalignment="center",
                        color="white" if cm[i, j] > thresh else "black")

        plt.tight_layout()
        plt.ylabel('Ground truth label')
        plt.xlabel('Predicted label - accuracy={:0.4f}; misclass={:0.4f}'.format(accuracy, misclass))
        # plt.savefig(path)
        plt.tight_layout(pad=0.4, h_pad=0.4, w_pad=0.4)
        plt.savefig(path, dpi=500)

    def exportExperimentResults(self, fName, cmc_results_np: np.ndarray):
        """
        Method to export cmc results as csv file.

        :param fName: [required] name of result file.
        :param cmc_results_np: numpy array holding the camera movements classification predictions for each shot of a movie.
        """

        print("export results to csv!")

        if (len(cmc_results_np) == 0):
            print("ERROR: numpy is empty")
            exit()

        print(self.config_instance.path_eval_results)
        fp = open(self.config_instance.path_eval_results + "/" + fName.split('/')[-1].split('.')[0] + ".csv", 'w')

        header = "exp_name;mvi_mv_ratio;threshold_significance;threshold_consistency;mvi_window_size;region_window_size;active_threshold;acc;prec;rec;f1_score"
        fp.write(header + "\n")

        for i in range(0, len(cmc_results_np)):
            tmp_line = str(cmc_results_np[i][0])
            for c in range(1, len(cmc_results_np[i])):
                tmp_line = tmp_line + ";" + str(cmc_results_np[i][c])
            fp.write(tmp_line + "\n")

        fp.close()



from tracemalloc import start
import cv2
import numpy as np
import argparse
from matplotlib import pyplot as plt
import matplotlib.gridspec as gridspec
from vhh_cmc.OpticalFlow_ORB import OpticalFlow_ORB
from vhh_cmc.OpticalFlow_SIFT import OpticalFlow_SIFT
from vhh_cmc.OpticalFlow_Dense import OpticalFlow_Dense
from datetime import datetime
import pymp
import math
from scipy.stats import multivariate_normal


class OpticalFlow(object):
    def __init__(self, video_frames=None, algorithm=None, config_instance=None):
        self.video_frames = video_frames
        self.config_instance = config_instance

        self.number_of_blocks = 32

    def calculate_displacements_u_v(self):
        print("INFO: calculate dense optical flow for entire shot ...")
        frames_np = np.squeeze(self.video_frames)
        print(frames_np.shape)

        of_dense_instance = OpticalFlow_Dense()

        start_time1 = datetime.now()
        frm_u_np = pymp.shared.array((frames_np.shape[0] - 1, frames_np.shape[1], frames_np.shape[2]), dtype='float32')
        frm_v_np = pymp.shared.array((frames_np.shape[0] - 1, frames_np.shape[1], frames_np.shape[2]), dtype='float32')
        frm_mag_np = pymp.shared.array((frames_np.shape[0] - 1, frames_np.shape[1], frames_np.shape[2]), dtype='float32')
        frm_ang_np = pymp.shared.array((frames_np.shape[0] - 1, frames_np.shape[1], frames_np.shape[2]), dtype='float32')
        with pymp.Parallel(4) as p:
            #p.print(p.num_threads, p.thread_num)
            for i in p.range(0, len(frames_np)-1):
                curr_frame = frames_np[i]
                nxt_frame = frames_np[i + 1]

                curr_frame = cv2.equalizeHist(curr_frame)
                nxt_frame = cv2.equalizeHist(nxt_frame)
                kernel = np.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]])
                curr_frame = cv2.filter2D(curr_frame, -1, kernel)
                nxt_frame = cv2.filter2D(nxt_frame, -1, kernel)
                frm_mag_np[i], frm_ang_np[i], frm_u_np[i], frm_v_np[i] = of_dense_instance.getFlow(curr_frame, nxt_frame)

        print(frm_mag_np.shape)
        print(frm_ang_np.shape)
        print(frm_u_np.shape)
        print(frm_v_np.shape)

        end_time1 = datetime.now()
        time_elapsed1 = end_time1 - start_time1
        print(time_elapsed1)

        return frm_mag_np, frm_ang_np, frm_u_np, frm_v_np

    def create_macro_blocks(self, frm_u_np, frm_v_np, frm_mag_np, frm_ang_np, VISUALIZE_ACTIVE_FLAG=False):
        # split into macro blocks (mb) - calculate one representative vector for each macro block (mean)
        start_time2 = datetime.now()
        
        all_mb_u_np = pymp.shared.array((frm_u_np.shape[0], self.number_of_blocks*self.number_of_blocks), dtype='float32')
        all_mb_v_np = pymp.shared.array((frm_v_np.shape[0], self.number_of_blocks*self.number_of_blocks), dtype='float32')
        all_mb_mag_np = pymp.shared.array((frm_mag_np.shape[0], self.number_of_blocks*self.number_of_blocks), dtype='float32')
        all_mb_ang_np = pymp.shared.array((frm_ang_np.shape[0], self.number_of_blocks*self.number_of_blocks), dtype='float32')
        all_mb_pos_x_np = pymp.shared.array((frm_mag_np.shape[0], self.number_of_blocks*self.number_of_blocks), dtype='int')
        all_mb_pos_y_np = pymp.shared.array((frm_ang_np.shape[0], self.number_of_blocks*self.number_of_blocks), dtype='int')

        with pymp.Parallel(4) as p:
            #p.print(p.num_threads, p.thread_num)
            for i in p.range(0, len(frm_u_np)):
            #for i in range(0, len(frm_u_np)):
                frm_u = frm_u_np[i]
                frm_v = frm_v_np[i]
                frm_mag = frm_mag_np[i]
                frm_ang = frm_ang_np[i]

                frm_mb_pos_x_l = []
                frm_mb_pos_y_l = []
                frm_mb_u_l = []
                frm_mb_v_l = []
                frm_mb_mag_l = []
                frm_mb_ang_l = []
                for r in range(0, self.number_of_blocks):
                    for c in range(0, self.number_of_blocks):
                        # block
                        mb_u, (mb_center_x, mb_center_y) = self.getBlock(
                            frame=frm_u,
                            row=r,
                            col=c,
                            number_of_blocks=self.number_of_blocks)
                        mb_v, (mb_center_x, mb_center_y) = self.getBlock(
                            frame=frm_v,
                            row=r,
                            col=c,
                            number_of_blocks=self.number_of_blocks)
                        mb_mag, (mb_center_x, mb_center_y) = self.getBlock(
                            frame=frm_mag,
                            row=r,
                            col=c,
                            number_of_blocks=self.number_of_blocks)
                        mb_ang, (mb_center_x, mb_center_y) = self.getBlock(
                            frame=frm_ang,
                            row=r,
                            col=c,
                            number_of_blocks=self.number_of_blocks)

                        frm_mb_u_l.append(np.mean(mb_u))
                        frm_mb_v_l.append(np.mean(mb_v))
                        frm_mb_mag_l.append(np.mean(mb_mag))
                        frm_mb_ang_l.append(np.mean(mb_ang))
                        frm_mb_pos_x_l.append(mb_center_x)
                        frm_mb_pos_y_l.append(mb_center_y)

                all_mb_u_np[i] = np.array(frm_mb_u_l)
                all_mb_v_np[i] = np.array(frm_mb_v_l)
                all_mb_mag_np[i] = np.array(frm_mb_mag_l)
                all_mb_ang_np[i] = np.array(frm_mb_ang_l)
                all_mb_pos_x_np[i] = np.array(frm_mb_pos_x_l)
                all_mb_pos_y_np[i] = np.array(frm_mb_pos_y_l)

        print(all_mb_mag_np.shape)
        print(all_mb_ang_np.shape)
        print(all_mb_u_np.shape)
        print(all_mb_v_np.shape)
        print(all_mb_pos_x_np.shape)
        print(all_mb_pos_y_np.shape)
        
        end_time2 = datetime.now()
        time_elapsed2 = end_time2 - start_time2
        print(time_elapsed2)

        if(VISUALIZE_ACTIVE_FLAG == True):
            frames_np = np.squeeze(self.video_frames)
            print(frames_np.shape)
            self.visualize_motion_vectors(frames_np, all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np)

        return all_mb_mag_np, all_mb_ang_np, all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np

    def visualize_motion_vectors(self, frames_np, u_np, v_np, x_np, y_np, mask=None):
        # visualize vectors and frames
        for i in range(0, len(frames_np) - 1):
            frame_rgb = cv2.cvtColor(frames_np[i], cv2.COLOR_GRAY2RGB)
            for s in range(0, len(u_np[i]) - 2):
                if(mask is None):
                    point_color = (0, 0, 255)
                    line_color = (0, 0, 255)
                else:
                    # switch color
                    if (mask[i][s]):
                        point_color = (0, 0, 0)
                        line_color = (0, 255, 0)
                    else:
                        point_color = (0, 0, 255)
                        line_color = (0, 0, 255)
                cv2.circle(frame_rgb, center=(x_np[i][s], y_np[i][s]), radius=1, thickness=1,
                           color=point_color)
                cv2.line(frame_rgb, pt1=(x_np[i][s], y_np[i][s]),
                         pt2=(int(x_np[i][s] + u_np[i][s]), int(y_np[i][s] + v_np[i][s])),
                         thickness=1,
                         color=line_color)

            win_name = "orig+vectors"  # 1. use var to specify window name everywhere
            cv2.namedWindow(win_name, cv2.WINDOW_NORMAL)  # 2. use 'normal' flag
            cv2.imshow(win_name, frame_rgb)
            cv2.resizeWindow(win_name, frame_rgb.shape[0], frame_rgb.shape[1])
            s = cv2.waitKey(100)

    def filter_motion_vectors_of_interest_OLD(self, all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np, k, n, t1, t2, VISUALIZE_ACTIVE_FLAG=True):
        all_filter_masks_l = []
        all_seq_mb_delta_u_l = []
        all_seq_mb_delta_v_l = []
        last_conditions = np.zeros((self.number_of_blocks*self.number_of_blocks)).astype('bool')
        last_conditions.fill(False)

        for i in range(0, len(all_mb_u_np) - k):
            # calculate delta_u and delta_v - horizontal and vertical displacements
            seq_mb_delta_u = []
            seq_mb_delta_v = []
            for j in range(0, k):
                u_curr = all_mb_u_np[i + j]
                v_curr = all_mb_v_np[i + j]
                u_next = all_mb_u_np[i + j + 1]
                v_next = all_mb_v_np[i + j + 1]

                delta_u = u_next - u_curr
                delta_v = v_next - v_curr
                seq_mb_delta_u.append(delta_u)
                seq_mb_delta_v.append(delta_v)
            seq_mb_delta_u_np = np.array(seq_mb_delta_u)
            seq_mb_delta_v_np = np.array(seq_mb_delta_v)

            print("####### DELTA U-V #########")
            print(i)
            print(seq_mb_delta_u_np)
            print(seq_mb_delta_u_np.shape)
            print(seq_mb_delta_v_np)
            print(seq_mb_delta_v_np.shape)
            #continue

            # check significance and consistency
            seq_mb_mu_delta_u = np.mean(seq_mb_delta_u_np, axis=0)
            seq_mb_mu_delta_v = np.mean(seq_mb_delta_v_np, axis=0)

            all_seq_mb_delta_u_l.append(seq_mb_mu_delta_u)
            all_seq_mb_delta_v_l.append(seq_mb_mu_delta_v)

            sum_seq_mb_sigma_delta_u = np.square(seq_mb_delta_u_np[0] - seq_mb_mu_delta_u)
            sum_seq_mb_sigma_delta_v = np.square(seq_mb_delta_v_np[0] - seq_mb_mu_delta_v)
            for j in range(1, k):
                sum_seq_mb_sigma_delta_u = sum_seq_mb_sigma_delta_u + np.square(seq_mb_delta_u_np[j] - seq_mb_mu_delta_u)
                sum_seq_mb_sigma_delta_v = sum_seq_mb_sigma_delta_v + np.square(seq_mb_delta_v_np[j] - seq_mb_mu_delta_v)
            seq_mb_sigma_delta_u = np.sqrt((1 / (k - 1)) * sum_seq_mb_sigma_delta_u)
            seq_mb_sigma_delta_v = np.sqrt((1 / (k - 1)) * sum_seq_mb_sigma_delta_v)

            print("####### SIGMA U-V #########")
            print(i)
            print(seq_mb_sigma_delta_u)
            print(seq_mb_sigma_delta_u.shape)
            print(seq_mb_sigma_delta_v)
            print(seq_mb_sigma_delta_v.shape)

            sum_seq_mb_mu_dist = np.sqrt((np.square(all_mb_u_np[i]) + np.square(all_mb_v_np[i])))
            for j in range(1, k):
                sum_seq_mb_mu_dist = sum_seq_mb_mu_dist + np.sqrt((np.square(all_mb_u_np[i+j]) + np.square(all_mb_v_np[i+j])))
            seq_mb_mu_dist = (1 / k) * sum_seq_mb_mu_dist

            print("####### Mean Magnitude #########")
            print(i)
            print(seq_mb_mu_dist)
            print(seq_mb_mu_dist.shape)

            print("####### MIN/MAX values #########")
            print(np.min(seq_mb_mu_dist))
            print(np.max(seq_mb_mu_dist))
            print(np.min(np.sqrt(np.square(seq_mb_sigma_delta_u) + np.square(seq_mb_sigma_delta_v))))
            print(np.max(np.sqrt(np.square(seq_mb_sigma_delta_u) + np.square(seq_mb_sigma_delta_v))))

            significance_condition1 = seq_mb_mu_dist > t1
            consistency_condition2 = np.sqrt(np.square(seq_mb_sigma_delta_u) + np.square(seq_mb_sigma_delta_v)) < t2
            final_condition = np.logical_and(significance_condition1, consistency_condition2)

            print("####### Conditions #########")
            print(significance_condition1)
            print(significance_condition1.shape)
            print(np.unique(significance_condition1, return_counts=True))
            print(consistency_condition2)
            print(consistency_condition2.shape)
            print(np.unique(consistency_condition2, return_counts=True))
            print(final_condition)
            print(final_condition.shape)
            print(np.unique(final_condition, return_counts=True))
            ''''''
            all_filter_masks_l.append(final_condition)

        all_filter_masks_np = np.array(all_filter_masks_l)
        print(all_filter_masks_np.shape)
        all_seq_mb_delta_u_np = np.array(all_seq_mb_delta_u_l)
        print(all_seq_mb_delta_u_np.shape)
        all_seq_mb_delta_v_np = np.array(all_seq_mb_delta_v_l)
        print(all_seq_mb_delta_v_np.shape)

        if(VISUALIZE_ACTIVE_FLAG == True):
            frames_np = np.squeeze(self.video_frames)
            print(frames_np.shape)
            self.visualize_motion_vectors(frames_np, all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np, all_filter_masks_np)
        
        return all_filter_masks_np

    def filter_motion_vectors_of_interest(self, all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np, k, n, t1, t2, VISUALIZE_ACTIVE_FLAG=True):
        start_time3 = datetime.now()
        
        # calculate gradients of horizontal and vertical displacement (delta_u and delta_v) for each macro block (mb)
        all_mb_delta_u_l = []
        all_mb_delta_v_l = []
        for i in range(0, len(all_mb_u_np) - 1):
            u_curr = all_mb_u_np[i]
            v_curr = all_mb_v_np[i]
            u_next = all_mb_u_np[i + 1]
            v_next = all_mb_v_np[i + 1]

            delta_u = u_next - u_curr
            delta_v = v_next - v_curr
            all_mb_delta_u_l.append(delta_u)
            all_mb_delta_v_l.append(delta_v)
        all_mb_delta_u_np = np.array(all_mb_delta_u_l)
        all_mb_delta_v_np = np.array(all_mb_delta_v_l)

        print("####### DELTA U-V #########")
        print(all_mb_delta_u_np.shape)
        print(all_mb_delta_v_np.shape)


        # calculate mean magnitude for significance check
        
        all_mu_dist_l = []
        all_sigma_delta_u_l = []
        all_sigma_delta_v_l = []
        all_mu_delta_u_l = []
        all_mu_delta_v_l = []
        all_val_l = []
        for i in range(0, len(all_mb_delta_u_np) - k): 
            #print("significance check")
            a = np.square(all_mb_u_np[i:i+k, :])
            b = np.square(all_mb_v_np[i:i+k, :])
            c = np.sqrt(a + b)
            
            mu_dist_np = np.sum(c, axis=0) / k
            all_mu_dist_l.append(mu_dist_np)

            #print("consistence check")
            val_l = []
            sigma_delta_u_l = []
            sigma_delta_v_l = []
            for b in range(0, all_mb_delta_u_np.shape[1]):
                mu_mb_delta_u = np.mean(all_mb_delta_u_np[i:i+k, b])
                mu_mb_delta_v = np.mean(all_mb_delta_v_np[i:i+k, b])
                mu = np.array([mu_mb_delta_u, mu_mb_delta_v])
                
                t_mb_delta_u = np.expand_dims(all_mb_delta_u_np[i:i+k, b], axis=0)
                t_mb_delta_v = np.expand_dims(all_mb_delta_v_np[i:i+k, b], axis=0)
                X = np.concatenate((t_mb_delta_u,t_mb_delta_v), axis=0)
                cov = np.cov(X)
                #print(cov)
                #print(cov.shape)
           
                sigma_delta_u_np = cov[0][0]
                sigma_delta_v_np = cov[1][1]
                #print(sigma_delta_u_np)
                #print(sigma_delta_v_np)
                
                sigma_delta_u_l.append(sigma_delta_u_np)
                sigma_delta_v_l.append(sigma_delta_v_np)

                # sigma values
                val = np.sqrt(np.square(sigma_delta_u_np) + np.square(sigma_delta_v_np))
                # multivariate 
                #rv = multivariate_normal(mean=mu, cov=cov)
                #val = rv.pdf(x=all_mb_delta_u_np[i, b])
                val_l.append(val)

            val_np = np.array(val_l)
            all_val_l.append(val_np)
            
            all_sigma_delta_u_l.append(sigma_delta_u_l)
            all_sigma_delta_v_l.append(sigma_delta_v_l)
            #all_mu_delta_u_l.append(mu_delta_u_np)
            #all_mu_delta_v_l.append(mu_delta_v_np)


        all_mu_dist_np = np.array(all_mu_dist_l)
        print(all_mu_dist_np.shape)

        all_val_np = np.array(all_val_l)
        print(all_val_np.shape)
        print(all_val_np)
        print(np.min(all_val_np))
        print(np.max(all_val_np))
        print(np.mean(all_val_np))
        #exit()

        all_sigma_delta_u_np = np.array(all_sigma_delta_u_l)
        all_sigma_delta_v_np = np.array(all_sigma_delta_v_l)
        all_sigma_delta_u_np = np.nan_to_num(all_sigma_delta_u_np)
        all_sigma_delta_v_np = np.nan_to_num(all_sigma_delta_v_np)
        #all_mu_delta_u_np = np.array(all_mu_delta_u_l)
        #all_mu_delta_v_np = np.array(all_mu_delta_v_l)
        print(all_sigma_delta_u_np.shape)
        print(all_sigma_delta_v_np.shape)
        #print(all_mu_delta_u_np.shape)
        #print(all_mu_delta_v_np.shape)

        # condition1 significance check
        mask1 = all_mu_dist_np > t1
        print(np.min(all_mu_dist_np))
        print(np.max(all_mu_dist_np))
        print(np.mean(all_mu_dist_np))

        # condition2 consistence check
        mask2 = all_val_np < t2
        print(mask2)
        print(mask2.shape)

        # filter mask for mvi
        mvi_mask = np.logical_and(mask1, mask2)
        print(mvi_mask)

        if(VISUALIZE_ACTIVE_FLAG == True):
            frames_np = np.squeeze(self.video_frames)
            print(frames_np.shape)
            self.visualize_motion_vectors(frames_np, all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np, mvi_mask)
        
        return mvi_mask

    def visualize_region_results(self, frames_np, all_motion_np, center_pos):
        # visualize region results vectors and frames
        for i in range(0, len(frames_np) - 1):
            frame_rgb = cv2.cvtColor(frames_np[i], cv2.COLOR_GRAY2RGB)
            for s in range(0, len(all_motion_np[i])):

                font = cv2.FONT_HERSHEY_SIMPLEX
                org = (center_pos[s][0] + int(170/2), center_pos[s][1] + int(170/2))
                fontScale = 1
                color = (255, 0, 0)
                thickness = 2
                frame_rgb = cv2.putText(frame_rgb, all_motion_np[i][s], org, font, 
                                fontScale, color, thickness, cv2.LINE_AA)

            win_name = "orig+mv"  # 1. use var to specify window name everywhere
            cv2.namedWindow(win_name, cv2.WINDOW_NORMAL)  # 2. use 'normal' flag
            cv2.imshow(win_name, frame_rgb)
            cv2.resizeWindow(win_name, frame_rgb.shape[0], frame_rgb.shape[1])
            s = cv2.waitKey(100)

    def region_estimation(self, all_mb_u_np, all_mb_v_np, all_mb_mag_np, all_mb_ang_np, mvi_mask, n, k, mvi_mv_ratio, VISUALIZE_ACTIVE_FLAG=False):
        all_mb_u_np = all_mb_u_np[:-1, :]
        all_mb_v_np = all_mb_v_np[:-1, :]
        all_mb_ang_np = all_mb_ang_np[:-1, :]

        print(all_mb_u_np.shape)
        print(mvi_mask.shape)
        print(all_mb_ang_np.shape)
        print(all_mb_mag_np.shape)
        
        blocks = [[0,113], [113,226], [226,339], [339,452], [452,565], [565,678], [678,791], [791,904], [904,1024]]
        vis_blocks = [[0,0], [0,170], [0,340], [170,0], [170,170], [170,340], [340,0], [340,170], [340,340]]
        
        motion_l = []
        all_pc_list = []
        for j in range(0, len(all_mb_u_np) - n - k):
            print("###### result per frame ######")
            local_motion_l = []
            pc_list = []
            for block in blocks:
                #print(block)
                start = block[0]
                stop = block[1]
            
                mb_u_region_bl_np = all_mb_u_np[j:j+n, start:stop]
                mb_v_region_bl_np = all_mb_v_np[j:j+n, start:stop]
                mb_ang_region_bl_np = all_mb_ang_np[j:j+n, start:stop]
                mb_mag_region_bl_np = all_mb_mag_np[j:j+n, start:stop]
                mvi_mask_region_bl_np = mvi_mask[j:j+n, start:stop]

                #mag_mvi = np.mean(np.abs(mb_mag_region_bl_np[mvi_mask_region_bl_np == True]))
                #mag_mv = np.mean(np.abs(mb_mag_region_bl_np))
                #final_condition = mag_mvi < mag_mv * 0.1

                u_res1 = np.sum(np.abs(mb_u_region_bl_np[mvi_mask_region_bl_np == True]))
                u_res2 = np.sum(np.abs(mb_u_region_bl_np ))
                v_res1 = np.sum(np.abs(mb_v_region_bl_np[mvi_mask_region_bl_np == True]))
                v_res2 = np.sum(np.abs(mb_v_region_bl_np ))
                condition_part_a = u_res1 < u_res2 * mvi_mv_ratio
                condition_part_b = v_res1 < v_res2 * mvi_mv_ratio
                final_condition = condition_part_a and condition_part_b
                ''''''
                
                if(final_condition == True):
                    print("NA")
                    local_motion_l.append("NA")
                    pc_list.append(0)
                else:
                    #print("other movement")
                    #local_motion_l.append("other movement")

                    u_mvis = mb_u_region_bl_np[mvi_mask_region_bl_np == True]
                    v_mvis = mb_v_region_bl_np[mvi_mask_region_bl_np == True]
                    ang_np = mb_ang_region_bl_np[mvi_mask_region_bl_np == True]
                    #print(ang_np)

                    u_mvis = np.expand_dims(u_mvis, axis=1)
                    v_mvis = np.expand_dims(v_mvis, axis=1)

                    A = np.concatenate((u_mvis, v_mvis), axis=1)
                    #print(A .shape)
                    u, s, vh = np.linalg.svd(A)
                    #print(f'u: {u}\n')
                    #print(f's: {s}\n')
                    #print(f'vh: {vh}\n')

                    pc_of_region = vh[:,:1].squeeze()
                    #print(f'pc_of_region: {pc_of_region}')

                    # range [-pi, pi]
                    #theta_region = np.arctan2(pc_of_region[1], pc_of_region[0])
                    #theta_region = np.degrees(theta_region)
                    if(pc_of_region[1] >= 0):
                        theta_region = np.degrees(np.arccos(np.dot([1, 0], pc_of_region.T)))
                    else:
                        theta_region = 360 - np.degrees(np.arccos(np.dot([1, 0], pc_of_region.T)))

                    
                    #print(f'theta_region: {theta_region}')


                    #print(f'theta_region: {np.arctan2(pc_of_region[1], pc_of_region[0])}')
                    #theta_region_degrees = np.degrees(theta_region)
                    pc_list.append(theta_region)


                    if ((theta_region > 45 and theta_region < 135) or (theta_region > 225 and theta_region < 315)):
                        class_name = "TILT"
                    elif ((theta_region > 135 and theta_region < 225) or
                          (theta_region > 315 and theta_region < 360) or
                          (theta_region > 0 and theta_region < 45)):
                        class_name = "PAN"
                    else:
                        class_name = "NA"
                    print(class_name)
                    print(theta_region)
                    local_motion_l.append(class_name)

                    #plt.figure()
                    #plt.scatter(u_mvis, v_mvis)
                    #plt.show()
                    #exit()
                    #print(f'theta_region(degrees): {theta_region_degrees}')
        
            motion_l.append(local_motion_l)
            all_pc_list.append(pc_list)
        
        all_motion_np = np.array(motion_l)
        print(all_motion_np)
        print(all_motion_np.shape)
        print(np.unique(all_motion_np, return_counts=True))

        print(np.array(all_pc_list))
        print(np.array(all_pc_list).shape)

        region_block = np.array(all_pc_list)
        print(region_block)
        print(region_block.shape)

        # visualize region results vectors and frames
        if(VISUALIZE_ACTIVE_FLAG == True):
            frames_np = np.squeeze(self.video_frames)
            self.visualize_region_results(frames_np, all_motion_np, vis_blocks)
        ''''''

        return all_motion_np


    def runDense(self, vid_name, shot_id, shot_start, shot_end):
        frames_np = np.squeeze(self.video_frames)
        print(frames_np.shape)

        if(len(frames_np) > 10000):
            print("WARNING: shot is very very long! --> skip")
            class_name = "NA"
            return class_name, []

        #calculate dense optical flow vectors (u, v, mag, ang)
        frm_mag_np, frm_ang_np, frm_u_np, frm_v_np = self.calculate_displacements_u_v()

        # split into macro blocks (mb)
        all_mb_mag_np, all_mb_ang_np, all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np = self.create_macro_blocks(frm_u_np, frm_v_np, frm_mag_np, frm_ang_np)
        
        # motion vector of interest detection (MVI)
        k = self.config_instance.mvi_window_size
        n = self.config_instance.region_window_size
        t1 = self.config_instance.threshold_significance
        t2 = self.config_instance.threshold_consistency
        mvi_mv_ratio = self.config_instance.mvi_mv_ratio

        #self.filter_motion_vectors_of_interest(all_mb_u_np, all_mb_v_np, k, n, t1, t2)
        mvi_mask = self.filter_motion_vectors_of_interest_OLD(all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np, k, n, t1, t2, VISUALIZE_ACTIVE_FLAG=False)
        #mvi_mask = self.filter_motion_vectors_of_interest(all_mb_u_np, all_mb_v_np, all_mb_pos_x_np, all_mb_pos_y_np, k, n, t1, t2, VISUALIZE_ACTIVE_FLAG=False)

        # region estimation
        all_motion_np = self.region_estimation(all_mb_u_np, all_mb_v_np, all_mb_mag_np, all_mb_ang_np, mvi_mask, n, k, mvi_mv_ratio, VISUALIZE_ACTIVE_FLAG=False)

        # final region assessment
        print("final region assessment")
        all_final_motion_l = []
        for r in range(0, len(all_motion_np)):
            cls, cnts = np.unique(all_motion_np[r, :], return_counts=True)
            idx = np.argmax(cnts)
            final_class_name = cls[idx]
            all_final_motion_l.append(final_class_name)
           
        all_final_motion_np = np.array(all_final_motion_l)
        print(all_final_motion_np.shape)
        print(all_final_motion_np)

        # separate pan and tilt list
        print("#################################")
        print("split into pan and tilt list ... ")
        pan_list = self.find_movement_in_sequence(find_class="PAN", motion_np=all_final_motion_np)
        tilt_list = self.find_movement_in_sequence(find_class="TILT", motion_np=all_final_motion_np)
        pans_np = np.array(pan_list)
        tilts_np = np.array(tilt_list)
        print(pans_np)
        print(tilts_np)
                
        # condition A --> remove short movements  
        condition_a_flag = True
        if(condition_a_flag == True):   
            print("########################################")  
            print("Condition A: remove short movements ... ")
            min_length_of_motion = 20 #10
            pans_np = self.filter_short_movements(pans_np, min_length_of_motion=min_length_of_motion)
            tilts_np = self.filter_short_movements(tilts_np, min_length_of_motion=min_length_of_motion)
            print(pans_np)
            print(tilts_np)
        else:
            print("#######################################################")  
            print("Condition A: remove short movements --> NOT ACTIVE ... ")
            print()

        # Condition B --> Find Gaps
        condition_b_flag = True
        if(condition_b_flag == True):   
            print("######################################")
            print("Condition B: find and filter gaps ... ")
            max_length_of_gap = 5 #2

            if (len(tilts_np) >= 2):
                tilts_np = self.filter_movements_gaps(tilts_np, max_length_of_gap=max_length_of_gap)
                print(tilts_np)

            if (len(pans_np) >= 2):
                pans_np = self.filter_movements_gaps(pans_np, max_length_of_gap=max_length_of_gap)
                print(pans_np)
        else:
            print("#######################################################")  
            print("Condition B: find and filter gaps --> NOT ACTIVE ... ")
            print()

        # map movment lists to shot boundaries
        final_movements_np = None
        if(len(tilts_np) > 0 and len(pans_np) > 0):
            final_movements_np = np.concatenate((tilts_np, pans_np), axis=0)
        elif(len(tilts_np) > 0 and len(pans_np) <= 0):
            final_movements_np = tilts_np
        elif(len(pans_np) > 0 and len(tilts_np) <= 0):
            final_movements_np = pans_np
        
        #exit()
        seq_dict_l = []
        if(final_movements_np is not None):    
            tmp1 = final_movements_np[:,:2].astype('int') + shot_start 
            tmp2 = final_movements_np[:,2:]
            final_movements_np = np.concatenate((tmp1, tmp2), axis=1)
            
            for g in range(0, len(final_movements_np)):
                seq_start = final_movements_np[g][0]
                seq_stop = final_movements_np[g][1]
                class_name = final_movements_np[g][2]
                #line = [vid_name.split('.')[0], shot_id, seq_start, seq_stop, class_name]
                seq_dict = {
                    "shotId": shot_id,
                    "start": seq_start,
                    "stop": seq_stop,
                    "cmcType": class_name
                }
                seq_dict_l.append(seq_dict)
        
            final_movements_duration_np = np.abs(final_movements_np[:,0:1].astype('int') - final_movements_np[:,1:2].astype('int'))
            final_movements_duration_np = np.concatenate((final_movements_duration_np, final_movements_np[:,2:]), axis=1)
            print(final_movements_duration_np)

            idx = np.where(final_movements_duration_np[:, 1:] == "PAN")[0]
            pan_duration = np.sum(final_movements_duration_np[idx, :1].astype('int'))
            print(pan_duration)

            idx = np.where(final_movements_duration_np[:, 1:] == "TILT")[0]
            tilt_duration = np.sum(final_movements_duration_np[idx, :1].astype('int'))
            print(tilt_duration)

            if(pan_duration >= tilt_duration): 
                class_name = "PAN"
            elif(tilt_duration > pan_duration): 
                class_name = "TILT"
            else:
                class_name = "NA"
            print(class_name)
        else:
            class_name = "NA"
            print(class_name)

        #exit()
        return class_name, seq_dict_l

    def window_filter(self, data_np, window_size=10, alpha=1.5):
        data_mean_l = []
        center = int(window_size/2)
        for i in range(0, len(data_np)):
            if (i-center < 0):
                filtered = np.nan_to_num(np.mean(data_np[i:i+center]))
            else:
                filtered = np.nan_to_num(np.mean(data_np[i-center:i+center]))
            data_mean_l.append(filtered)
        data_mean_np = np.array(data_mean_l)
        return data_mean_np

    def outlier_filter(self, data_np, alpha=1.0):
        # print(type(data_np))
        # print(data_np.shape)
        data_std = np.std(data_np)
        data_mean = np.mean(data_np)
        anomaly_cut_off = data_std * alpha
        lower_limit = data_mean - anomaly_cut_off
        upper_limit = data_mean + anomaly_cut_off

        # Generate outliers
        outliers_idx = []
        filtered_data = []
        for j in range(0, len(data_np)):
            if (j < len(data_np) - 1):
                nxt = data_np[j + 1]
                curr = data_np[j]
                prv = data_np[j - 1]
            else:
                nxt = data_np[j]
                curr = data_np[j]
                prv = data_np[j - 1]

            if curr > upper_limit or curr < lower_limit:
                data_mean = (prv + nxt) / 2.0
                filtered_data.append(data_mean)
                outliers_idx.append(j)
            else:
                filtered_data.append(curr)

        filtered_data_np = np.array(filtered_data)
        return filtered_data_np, outliers_idx

    def getBlock(self, frame=None, row=-1, col=-1, number_of_blocks=4):
        frame_w = frame.shape[1]
        frame_h = frame.shape[0]

        block_w = int(frame_w / number_of_blocks)
        block_h = int(frame_h / number_of_blocks)

        start_idx_row = row * block_h
        start_idx_col = col * block_w
        stop_idx_row = row * block_h + block_h
        stop_idx_col = col * block_w + block_w

        block_center_x = start_idx_col + int(abs(start_idx_col - stop_idx_col) / 2)
        block_center_y = start_idx_row + int(abs(start_idx_row - stop_idx_row) / 2)

        if (len(frame.shape) == 3):
            frame_block = frame[start_idx_row:stop_idx_row, start_idx_col:stop_idx_col, :]
        elif (len(frame.shape) == 2):
            frame_block = frame[start_idx_row:stop_idx_row, start_idx_col:stop_idx_col]
        else:
            print("ERROR: something is wrong with the frame shape.")

        return frame_block, (block_center_x, block_center_y)

    def find_movement_in_sequence(self, find_class="TILT", motion_np=None):
        state = "start"
        start = 0
        stop = len(motion_np)
        class_name_prev = "NA"
        movement_list = []

        unique_cmcs = np.unique(motion_np)
        if(len(unique_cmcs) == 1 and unique_cmcs[0] == find_class):
            movement_list.append([start, stop, unique_cmcs[0]])
        else:
            for h in range(0, len(motion_np)):
                #print(h)
                class_name = motion_np[h]

                # find start of movement
                if(class_name == find_class and state == "start"):
                    #print("state: start")
                    start = h
                    state = "stop"
                    
                elif(class_name_prev != class_name and state == "stop"):
                    #print("state: stop")
                    stop = h-1
                    movement_list.append([start, stop, class_name_prev])
                    state = "start"
                elif(h == (len(motion_np) - 1) and state == "stop"):
                    #print("end of list")
                    stop = len(motion_np) - 1 
                    movement_list.append([start, stop, class_name_prev])
                class_name_prev = class_name
        return movement_list

    def filter_short_movements(self, motion_np, min_length_of_motion=5):
        filtered_motion_list = []
        for h in range(0, len(motion_np)):
            start = int(motion_np[h][0])
            stop = int(motion_np[h][1])
            class_name = motion_np[h][2]
            if(stop < start):
                print("ERROR: something is wrong(DEBUG Point C)!!")
                exit()

            difference = stop - start
            if(difference >= min_length_of_motion):
                filtered_motion_list.append([start, stop, class_name])

        return np.array(filtered_motion_list)

    def filter_movements_gaps(self, motion_np, max_length_of_gap=5):
        m_l = []

        if (len(motion_np) <= 0):
            return np.array(m_l)

        start_seed = motion_np[0][0]
        stop_seed = motion_np[0][1]
        class_seed = motion_np[0][2]

        for h in range(1, len(motion_np)):
            start_nxt = motion_np[h][0]
            stop_nxt = motion_np[h][1]
            class_nxt = motion_np[h][2]

            difference = abs(int(stop_seed) - int(start_nxt))
            if(difference <= max_length_of_gap):
                stop_seed = stop_nxt
            else:
                m_l.append([start_seed, stop_seed, class_seed])
                start_seed = start_nxt
                stop_seed = stop_nxt
                class_seed = class_nxt

            if (h >= len(motion_np) - 1):
                m_l.append([start_seed, stop_seed, class_seed])
        return np.array(m_l)

    def filter1D(self, data_np, alpha=2.0):
        # print(type(data_np))
        # print(data_np.shape)
        data_std = np.std(data_np)
        data_mean = np.mean(data_np)
        anomaly_cut_off = data_std * alpha
        lower_limit = data_mean - anomaly_cut_off
        upper_limit = data_mean + anomaly_cut_off

        # Generate outliers
        outliers_idx = []
        filtered_data = []
        for j in range(0, len(data_np)):
            if (j < len(data_np) - 1):
                nxt = data_np[j + 1]
                curr = data_np[j]
                prv = data_np[j - 1]
            else:
                nxt = data_np[j]
                curr = data_np[j]
                prv = data_np[j - 1]

            if curr > upper_limit or curr < lower_limit:
                data_mean = (prv + nxt) / 2.0
                filtered_data.append(data_mean)
                outliers_idx.append(j)
            else:
                filtered_data.append(curr)

        filtered_data_np = np.array(filtered_data)
        return filtered_data_np, outliers_idx



import cv2
import numpy as np
from vhh_cmc.Configuration import Configuration


class PreProcessing(object):
    """
    This class is used to pre-process frames.
    """

    def __init__(self, config_instance: Configuration):
        """
        Constructor

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                            must be with extension ".yaml"
        """
        print("create instance of pre-processing ... ")
        self.config_instance = config_instance

    def applyTransformOnImg(self, image: np.ndarray) -> np.ndarray:
        """
        This method is used to apply the configured pre-processing methods on a numpy frame.

        :param image: This parameter must hold a valid numpy image (WxHxC).
        :return: This methods returns the preprocessed numpy image.
        """
        image_trans = image

        # convert to grayscale image
        if(int(self.config_instance.flag_convert2Gray) == 1):
            image_trans = self.convertRGB2Gray(image_trans)

        # crop image
        if (int(self.config_instance.center_crop_flag) == 1):
            image_trans = self.crop(image_trans, (image_trans.shape[0], image_trans.shape[0]))
            #image_trans = self.crop(image_trans, (700, 512))

        # resize image
        if(self.config_instance.flag_downscale == 1):
            dim = self.config_instance.resize_dim
            image_trans = self.resize(image_trans, dim)

        return image_trans

    def convertRGB2Gray(self, img: np.ndarray):
        """
        This method is used to convert a RBG numpy image to a grayscale image.

        :param img: This parameter must hold a valid numpy image.
        :return: This method returns a grayscale image (WxHx1).
        """
        img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        #img_gray = np.expand_dims(img_gray, axis=-1)
        return img_gray

    def crop(self, img: np.ndarray, dim: tuple):
        """
        This method is used to crop a specified region of interest from a given image.

        :param img: This parameter must hold a valid numpy image.
        :param dim: This parameter must hold a valid tuple including the crop dimensions.
        :return: This method returns the cropped image.
        """
        crop_w, crop_h = dim

        crop_h_1 = 0
        crop_h_2 = 0
        crop_w_1 = 0
        crop_w_2 = 0

        img_h = img.shape[0]
        img_w = img.shape[1]

        crop_w_1 = int(img_w / 2) - int(crop_w / 2)
        if (crop_w_1 < 0):
            crop_w_1 = 0

        crop_w_2 = int(img_w / 2) + int(crop_w / 2)
        if (crop_w_2 >= img_w):
            crop_w_2 = img_w

        crop_h_1 = int(img_h / 2) - int(crop_h / 2)
        if (crop_h_1 < 0):
            crop_h_1 = 0

        crop_h_2 = int(img_h / 2) + int(crop_h / 2)
        if (crop_h_2 >= img_h):
            crop_h_2 = img_h

        img_crop = img[crop_h_1:crop_h_2, crop_w_1:crop_w_2]
        return img_crop

    def resize(self, img: np.ndarray, dim: tuple):
        """
        This method is used to resize a image.

        :param img: This parameter must hold a valid numpy image.
        :param dim: This parameter must hold a valid tuple including the resize dimensions.
        :return: This method returns the resized image.
        """
        img_resized = cv2.resize(img, dim)
        if(int(self.config_instance.flag_convert2Gray) == 1):
            img_resized = np.expand_dims(img_resized, axis=-1)
        return img_resized





import cv2
import numpy as np
import argparse
from matplotlib import pyplot as plt


class OpticalFlow_Dense(object):
    def __init__(self):
        print("dense of")

    def getFlow(self, frame1, frame2):
        flow = cv2.calcOpticalFlowFarneback(frame1,
                                            frame2,
                                            None,
                                            pyr_scale=0.5,
                                            levels=3,
                                            winsize=15,
                                            iterations=3,  # 3
                                            poly_n=5,  # 5
                                            poly_sigma=1.2,
                                            flags=0)
        mag, ang = cv2.cartToPolar(flow[...,0], flow[...,1], angleInDegrees=True)

        '''
        hsv = np.zeros((frame1.shape[0],frame1.shape[1], 3), dtype=np.uint8)
        print(hsv.shape)
        print(frame1.shape)
        hsv[..., 1] = 255

        hsv[..., 0] = ang * 180 / np.pi / 2
        hsv[..., 2] = cv2.normalize(mag, None, 0, 255, cv2.NORM_MINMAX)
        bgr = cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)
        cv2.imshow("colored flow", bgr)
        cv2.imshow("orig", frame1)
        cv2.waitKey(10)
        '''
        return mag, ang, flow[...,0], flow[...,1]


import cv2
import numpy as np
import argparse
from matplotlib import pyplot as plt


class OpticalFlow_ORB(object):
    def __init__(self, video_frames=None):
        self.video_frames = video_frames

        print(self.video_frames[0].shape)
        frame_size = (self.video_frames[0].shape[1]*2, self.video_frames[0].shape[0])
        self.video_writer = cv2.VideoWriter("/data/share/output.avi", cv2.VideoWriter_fourcc(*"MJPG"), 12, frame_size)

    def getMatches(self, frame1, frame2, distance_threshold=0.75):
        kp_curr_list = []
        kp_prev_list = []

        orb = cv2.ORB_create(nfeatures=2000,
                             #scaleFactor=0.5,
                             #nlevels=3,
                             edgeThreshold=5,
                             #firstLevel=None,
                             WTA_K=3,
                             #scoreType=None,
                             #patchSize=None,
                             #fastThreshold=None
                             )
        kp_prev, descriptor_prev = orb.detectAndCompute(frame1, None)
        kp_curr, descriptor_curr = orb.detectAndCompute(frame2, None)

        #print(type(descriptor_prev))
        #print(type(descriptor_curr))
        '''
        out_img_prev = frame1.copy()
        #img = cv2.drawKeypoints(old_frame, kp, out_img)
        out_img_prev = cv2.drawKeypoints(frame1,
                                         kp_prev,
                                         out_img_prev,
                                         flags=cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)
        out_img_curr = frame2.copy()
        # img = cv2.drawKeypoints(old_frame, kp, out_img)
        out_img_curr = cv2.drawKeypoints(frame2,
                                         kp_curr,
                                         out_img_curr,
                                         flags=cv2.DRAW_MATCHES_FLAGS_DRAW_RICH_KEYPOINTS)
        #print(out_img_prev.shape)
        #print(out_img_curr.shape)
        '''

        ''''''
        #bf = cv2.BFMatcher()
        bf = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True)

        # Perform the matching between the ORB descriptors of the training image and the test image
        try:
            #matches = bf.knnMatch(descriptor_prev, descriptor_curr, k=2)
            matches = bf.match(descriptor_prev, descriptor_curr)
            #print(matches)
        except:
            return kp_prev_list, kp_curr_list

        # Sort them in the order of their distance.
        good_matches = sorted(matches, key=lambda x: x.distance)
        #print(good_matches)

        if (len(good_matches) <= 1):
            return kp_prev_list, kp_curr_list

        dist_l = [x.distance for x in good_matches]
        #dist_l.sort()
        #print(dist_l)

        x = np.array(dist_l)
        nom = (x - np.min(x)) / ((np.max(x) - np.min(x)) + 0.000000000001)
        dist_l_norm = np.nan_to_num(nom)
        #print(np.max(dist_l_norm))
        #print(np.min(dist_l_norm))

        idx = np.where(dist_l_norm < 0.15)[0]

        '''
        plt.plot(np.arange(len(dist_l_norm)), dist_l_norm)
        plt.draw()
        plt.pause(0.01)
        plt.cla()
        '''


        '''
        # Apply ratio test
        good_matches = []
        if (len(matches) > 0):
            if (len(matches[0]) == 2):
                for m, n in matches:
                    if m.distance < distance_threshold * n.distance:
                        good_matches.append([m])
            elif (len(matches[0]) == 1):
                for m in matches:
                    if m[0].distance < distance_threshold:
                        good_matches.append([m])
        '''

        # good_matches = np.squeeze(np.array(good_matches)).tolist()
        # print((good_matches))

        # print(matches)
        # print(type(good_matches))


        for match in good_matches[:np.max(idx)]:
            #kp_curr_list.append(kp_curr[match[0].trainIdx].pt)
            #kp_prev_list.append(kp_prev[match[0].queryIdx].pt)
            kp_curr_list.append(kp_curr[match.trainIdx].pt)
            kp_prev_list.append(kp_prev[match.queryIdx].pt)

        #print(frame1.shape)
        #print(frame2.shape)

        print(len(good_matches))
        result = cv2.drawMatches(frame1, kp_prev, frame2, kp_curr, good_matches[:np.max(idx)], outImg=None, flags=2)
        #result = cv2.drawMatches(out_img_prev, kp_prev, out_img_curr, kp_prev, good_matches, None, flags=2)
        # Display the best matching points
        #plt.title('Best Matching Points')
        '''
        plt.imshow(result)
        plt.draw()
        plt.pause(0.01)
        plt.cla()
        #self.video_writer.write(result)
        '''

        # Draw first 10 matches.
        #img3 = cv2.drawMatches(img1, kp1, img2, kp2, matches[:10], flags=2)

        return kp_prev_list, kp_curr_list



import yaml


class Configuration:
    """
    This class is needed to read the configuration parameters specified in the configuration.yaml file.
    The instance of the class is holding all parameters during runtime.

    .. note::
       e.g. ./config/config_vhh_test.yaml

        the yaml file is separated in multiple sections
        config['Development']
        config['PreProcessing']
        config['CmcCore']
        config['Evaluation']

        whereas each section should hold related and meaningful parameters.
    """

    def __init__(self, config_file: str):
        """
        Constructor

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                                       must be with extension ".yaml"
        """
        print("create instance of configuration ... ")

        if(config_file.split('.')[-1] != "yaml"):
            print("Configuration file must have the extension .yaml!")

        self.config_file = config_file

        # developer_config section
        self.debug_flag = -1
        self.sbd_results_path = None
        self.save_debug_pkg_flag = -1
        self.path_debug_results = None

        # pre-processing section
        self.flag_convert2Gray = -1
        self.center_crop_flag = -1
        self.flag_downscale = -1
        self.resize_dim = None

        # optical flow section
        self.mvi_mv_ratio = -1
        self.threshold_significance = -1
        self.threshold_consistency = -1
        self.mvi_window_size = -1
        self.region_window_size = -1
        self.active_threshold = -1

        # stc_core_config section
        self.class_names = None

        self.save_raw_results = -1
        self.path_postfix_raw_results = None
        self.path_prefix_raw_results = None
        self.path_raw_results = None

        self.save_final_results = -1
        self.path_prefix_final_results = None
        self.path_postfix_final_results = None
        self.path_final_results = None

        self.path_videos = None

        # evaluation section
        self.path_eval_dataset = None
        self.path_eval_results = ""
        self.save_eval_results = -1
        self.path_gt_data = None

    def loadConfig(self):
        """
        Method to load configurables from the specified configuration file
        """

        fp = open(self.config_file, 'r')
        config = yaml.load(fp, Loader=yaml.BaseLoader)

        developer_config = config['Development']
        pre_processing_config = config['PreProcessing']
        optical_flow_config = config['OpticalFlow']
        cmc_core_config = config['CmcCore']
        evaluation_config = config['Evaluation']

        # developer_config section
        self.debug_flag = int(developer_config['DEBUG_FLAG'])
        self.sbd_results_path = developer_config['SBD_RESULTS_PATH']
        self.save_debug_pkg_flag = int(developer_config['SAVE_DEBUG_PKG'])
        self.path_debug_results = developer_config['PATH_DEBUG_RESULTS']

        # pre-processing section
        self.flag_convert2Gray = int(pre_processing_config['CONVERT2GRAY_FLAG'])
        self.center_crop_flag = int(pre_processing_config['CENTER_CROP_FLAG'])
        self.flag_downscale = int(pre_processing_config['DOWNSCALE_FLAG'])
        self.resize_dim = (int(pre_processing_config['RESIZE_DIM'].split(',')[0]),
                           int(pre_processing_config['RESIZE_DIM'].split(',')[1]))

        # optical flow section
        self.mvi_mv_ratio = float(optical_flow_config['MVI_MV_RATIO'])
        self.threshold_significance = float(optical_flow_config['THRESHOLD_SIGNIFICANCE'])
        self.threshold_consistency = float(optical_flow_config['THRESHOLD_CONSISTENCY'])
        self.mvi_window_size = int(optical_flow_config['MVI_WINDOW_SIZE'])
        self.region_window_size = int(optical_flow_config['REGION_WINDOW_SIZE'])
        self.active_threshold = float(optical_flow_config['ACTIVE_THRESHOLD'])

        # cmc_core_config section
        self.class_names = cmc_core_config['CLASS_NAMES']

        self.save_raw_results = int(cmc_core_config['SAVE_RAW_RESULTS'])
        self.path_postfix_raw_results = cmc_core_config['POSTFIX_RAW_RESULTS']
        self.path_prefix_raw_results = cmc_core_config['PREFIX_RAW_RESULTS']
        self.path_raw_results = cmc_core_config['PATH_RAW_RESULTS']

        self.save_final_results = int(cmc_core_config['SAVE_FINAL_RESULTS'])
        self.path_prefix_final_results = cmc_core_config['PREFIX_FINAL_RESULTS']
        self.path_postfix_final_results = cmc_core_config['POSTFIX_FINAL_RESULTS']
        self.path_final_results = cmc_core_config['PATH_FINAL_RESULTS']

        self.path_videos = cmc_core_config['PATH_VIDEOS']

        # evaluation section
        self.path_eval_dataset = evaluation_config['PATH_EVAL_DATASET']
        self.path_eval_results = evaluation_config['PATH_EVAL_RESULTS']
        self.save_eval_results = int(evaluation_config['SAVE_EVAL_RESULTS'])
        self.path_gt_data = evaluation_config['PATH_GT_ANNOTATIONS']
